# (C) Copyright [2020] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from __future__ import absolute_import

from enum import Enum

from requests.structures import CaseInsensitiveDict

from .base_resource import AbstractWaitableResourceController, AbstractResource
from hpecp.exceptions import ContainerPlatformClientException
from hpecp.base_resource import ResourceList
from hpecp.user import User

try:
    basestring
except NameError:
    basestring = str


class TenantStatus(Enum):
    """Bases: enum.Enum

    The statuses for Tenant

    **Note:**

    The integer values do not have a meaning outside of this library.
    The API uses a string identifier with the status name rather than an
    integer value.
    """

    ready = 1
    creating = 2
    updating = 3
    upgrading = 4
    deleting = 5
    error = 6
    warning = 7


class Tenant(AbstractResource):

    all_fields = [
        "id",
        "name",
        "description",
        "status",
        "tenant_type",
        "external_user_groups",
    ]

    default_display_fields = [
        "id",
        "name",
        "description",
        "status",
        "tenant_type",
    ]

    @property
    def id(self):
        return self.json["_links"]["self"]["href"]

    @property
    def status(self):
        return self.json["status"]

    @property
    def name(self):
        return self.json["label"]["name"]

    @property
    def description(self):
        return self.json["label"]["description"]

    @property
    def tenant_type(self):
        try:
            return self.json["tenant_type"]
        except KeyError:
            return ""

    @property
    def external_user_groups(self):
        try:
            return self.json["external_user_groups"]
        except KeyError:
            return ""


class TenantController(AbstractWaitableResourceController):
    """Class that users will interact with to work with tenants.

    An instance of this class is available in
    `client.ContainerPlatformClient` with the attribute name
    :py:attr:`tenant <.client.ContainerPlatformClient.tenant>`. The methods of
    this class can be invoked using `client.tenant.method()`. See the example
    below.

    Example
    -------
    >>> client = ContainerPlatformClient(...).create_session()
    >>> client.tenant.list()
    """

    base_resource_path = "/api/v1/tenant/"
    resource_class = Tenant
    resource_list_path = "tenants"
    status_class = TenantStatus
    status_fieldname = "status"

    def __init__(self, client):
        self.client = client

    def create(
        self, name=None, description=None, tenant_type=None, k8s_cluster=None
    ):

        assert (
            isinstance(name, basestring) and len(name) > 0
        ), "'name' must be provided and must be a string"
        assert description is None or isinstance(
            description, basestring
        ), "'description' if provided, must be a string"

        data = {
            "label": {"name": name},
            "tenant_type": tenant_type,
            "member_key_available": "all_admins",
            "k8s_cluster": k8s_cluster,
        }
        if description is not None:
            data["label"]["description"] = description

        response = self.client._request(
            url="/api/v1/tenant",
            http_method="post",
            data=data,
            description="tenant/create",
        )
        return CaseInsensitiveDict(response.headers)["Location"]

    def k8skubeconfig(self):
        """Retrieve the tenant kubeconfig.

        This requires the ContainerPlatformClient to be created with
        a 'tenant' parameter.

        Returns
        -------
        str
            Tenant KubeConfig

        Raises
        ------
        ContainerPlatformClientException
            This is raised if the ContainerPlatformClient was not created with
            a 'tenant' parameter.
        """
        if self.client.tenant_config is None:
            raise ContainerPlatformClientException(
                "'tenant' session is required, but client "
                "was not create with a 'tenant' argument."
            )
        response = self.client._request(
            url="/api/v2/k8skubeconfig/",
            http_method="get",
            description="tenant/k8skubeconfig",
        )
        return response.text

    def get_external_user_groups(self, tenant_id):
        return self.get(tenant_id).external_user_groups

    def delete_external_user_group(
        self, tenant_id, group,
    ):
        user_groups = self.get_external_user_groups(tenant_id)

        # if group exists already, remove it
        user_groups = [
            ug for ug in user_groups if ug["group"].lower() != group.lower()
        ]

        data = {"external_user_groups": user_groups}
        self.client._request(
            url=tenant_id + "?external_user_groups",
            http_method="put",
            data=data,
            description="tenant/add_external_user_groups",
        )

    def add_external_user_group(
        self, tenant_id, group, role_id,
    ):
        user_groups = self.get_external_user_groups(tenant_id)

        # if group exists already, remove it
        user_groups = [ug for ug in user_groups if ug["group"] != group]

        # add group
        user_groups.append({"group": group, "role": role_id})

        data = {"external_user_groups": user_groups}
        self.client._request(
            url=tenant_id + "?external_user_groups",
            http_method="put",
            data=data,
            description="tenant/add_external_user_groups",
        )

    def users(self, id):
        response = self.client._request(
            url=id + "?user", http_method="get", description="tenant/users",
        )
        return ResourceList(User, response.json()["_embedded"]["users"],)

    def assign_user_to_role(self, tenant_id, role_id, user_id):
        """Assign a user to a given role using the tenant.

        Parameters
        ----------
        tenant_id : str
            The tenant ID - format: '/api/v1/tenant/[0-9]+'
        role_id : str
            The role ID - format: '/api/v1/role/[0-9]+'
        user_id : str
            The role ID - format: '/api/v1/user/[0-9]+'

        Raises
        ------
        APIItemNotFoundException
        APIItemConflictException
        APIException
        """
        # Ensure that the tenant is valid and exists
        self.get(tenant_id)

        # Ensure that the role is valid and exists
        self.client.role.get(role_id)

        # Ensure that the user is valid and exists
        self.client.user.get(user_id)

        data = {"operation": "assign", "role": role_id, "user": user_id}
        url = tenant_id + "?user"

        # Make the request
        self.client._request(
            url=url,
            http_method="put",
            data=data,
            description="assign_user_to_role",
        )

    def revoke_user_from_role(self, tenant_id, role_id, user_id):
        # Ensure that the tenant is valid and exists
        self.get(tenant_id)

        # Ensure that teh role is valid and exists
        self.client.role.get(role_id)

        # Ensure that the user is valid and exists
        self.client.user.get(user_id)

        data = {"operation": "revoke", "role": role_id, "user": user_id}
        url = tenant_id + "?user"

        # Make the request
        self.client._request(
            url=url,
            http_method="put",
            data=data,
            description="revoke_user_from_role",
        )



# (C) Copyright [2020] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

"""Base classes for Controllers and Resources."""

import abc
import polling
import six

from tabulate import tabulate
from hpecp.exceptions import APIItemNotFoundException
from .logger import Logger
import urllib

_log = Logger.get_logger()


@six.add_metaclass(abc.ABCMeta)
class AbstractResourceController:
    """Base class for Resource Controllers."""

    def _get_base_resource_path(self):
        return self._base_resource_path

    def _set_base_resource_path(self, path):
        assert (
            path.endswith("/") is False
        ), "base resource path must not end with '/'"
        self._base_resource_path = path

    base_resource_path = abc.abstractproperty(
        _get_base_resource_path, _set_base_resource_path
    )
    """Declare the resource base path for the API resource.

    :getter: Returns the resource base path
    :setter: Sets the resource base path
    :type: str

    Example
    -------
    class K8sClusterController(AbstractResourceController):
        ...
        base_resource_path = "/api/v2/k8scluster"
    """

    def _get_resource_class(self):
        return self._resource_class

    def _set_resource_class(self, clazz):
        self._resource_class = clazz

    resource_class = abc.abstractproperty(
        _get_resource_class, _set_resource_class
    )
    """Declare the implementing Resource class for the API resource.
    The resource class contains properties mapping to attributes in the
    response.

    :getter: Returns the Resource class
    :setter: Sets the Resource class
    :type: class

    Example
    -------
    class K8sClusterController(AbstractResourceController):
        ...
        resource_class = K8sCluster
    """

    def _get_resource_list_path(self):
        return self._resource_list_path

    def _set_resource_list_path(self, resource_list_path):
        self._resource_list_path = resource_list_path

    resource_list_path = abc.abstractproperty(
        _get_resource_list_path, _set_resource_list_path
    )
    """Declare the implementing resource list path for the API resource.
    The resource list path is where the resources are after the
    '_embedded' element in the API response json.

    :getter: Returns the resource list path
    :setter: Sets the resource list path
    :type: str

    Example
    -------
    class K8sClusterController(AbstractResourceController):
        ...
        resource_list_path = "k8sclusters"
    """

    def __init__(self, client):
        """Create a new instance.

        Parameters
        ----------
        client : ContainerPlatformClient
            client instance for working with the HPE CP API.
        """
        self.client = client

    def get(self, id, params={}):
        """Make an API call to retrieve a Resource.

        Parameters
        ----------
        id : str
            The ID with the format /resource/path/id
        params : str, optional
            API Parameters.

        Returns
        -------
        Instance of self.resource_class
            An instance of the class defined by the property
            self.resource_class

        Raises
        ------
        APIException
            The remote API returned an error.
        APIItemNotFoundException
            The item with {id} was not found.
        """
        assert isinstance(id, str), "'id' must be provided and must be a str"
        assert id.startswith(
            self.base_resource_path
        ), "'id' does not start with '{}'".format(self.base_resource_path)

        if params:
            if six.PY2:
                p = "?" + urllib.urlencode(params)
            else:
                p = "?" + urllib.parse.urlencode(params)
        else:
            p = ""

        response = self.client._request(
            url="{}{}".format(id, p),
            http_method="get",
            description=self.__class__.__name__ + "/get",
        )
        return self.resource_class(response.json())

    def list(self):
        """Make an API call to retrieve a list of Resources.

        Returns
        -------
        ResourceList
            The ResourceList will contain instances of the class defined by
            the property self.resource_class
        """
        response = self.client._request(
            url=self.base_resource_path,
            http_method="get",
            description=self.__class__.__name__ + "/list",
        )
        return ResourceList(
            self.resource_class,
            response.json()["_embedded"][self.resource_list_path],
        )

    def delete(self, id):
        """Make an API call to delete a Resources.

        Parameters
        ----------
        id : str
            The ID with the format /resource/path/id

        Raises
        ------
        APIException
            The remote API returned an error.
        APIItemNotFoundException
            The item with {id} was not found.
        """
        assert isinstance(id, str), "'id' must be provided and must be a str"
        assert id.startswith(
            self.base_resource_path
        ), "'id' does not start with '{}'".format(self.base_resource_path)

        self.client._request(
            url=id,
            http_method="delete",
            description=self.__class__.__name__ + "/delete",
        )


@six.add_metaclass(abc.ABCMeta)
class AbstractWaitableResourceController(AbstractResourceController):
    """Resource Controller that is able to wait for the resource's status."""

    def _get_status_class(self):
        return self._status_class

    def _set_status_class(self, clazz):
        self._status_class = clazz

    status_class = abc.abstractproperty(_get_status_class, _set_status_class)
    """Declare the implementing Status class for the API resource.
    The status class contains properties mapping to attributes in the
    response.

    :getter: Returns the Status class
    :setter: Sets the Status class
    :type: class

    Example
    -------
    class K8sClusterController(AbstractResourceController):
        ...
        status_class = K8sClusterStatus
    """

    def _get_status_fieldname(self):
        return self._status_fieldname

    def _set_status_fieldname(self, fieldname):
        self._status_fieldname = fieldname

    status_fieldname = abc.abstractproperty(
        _get_status_fieldname, _set_status_fieldname
    )
    """Declare the Status fieldname in the API resource.

    Usually either: status or state

    :getter: Returns the Status fieldname
    :setter: Sets the Status fieldname
    :type: str

    Example
    -------
    class K8sClusterController(AbstractResourceController):
        ...
        status_fieldname = status
    """

    def wait_for_state(self, id, states=[], timeout_secs=1200):
        """See wait_for_status()."""
        return self.wait_for_status(id, states, timeout_secs)

    def wait_for_status(self, id, status=[], timeout_secs=1200):
        """Wait for K8S worker status.

        Parameters
        ----------
        id: str
            The resource ID - format: '/resource/path/[0-9]+'
        status: list[:py:method:`status_class`]
            Status(es) to wait for.  Use an empty array if you want to
            wait for the resource existence to cease.
        timeout_secs: int
            How long to wait for the status(es) before raising an
            exception.

        Returns
        -------
        bool
            True if status was found before timeout, otherwise False
            True if item does not exist before timeout and status is empty
        """
        assert isinstance(status, list), "'status' must be a list"
        for i, s in enumerate(status):
            assert isinstance(
                s, self.status_class
            ), "'status' item '{}' is not of type {}".format(
                i, self.status_class
            )
        assert isinstance(timeout_secs, int), "'timeout_secs' must be an int"
        assert timeout_secs >= 0, "'timeout_secs' must be >= 0"

        # if status is empty return success when resource id not found
        if len(status) == 0:
            _log.debug(
                "waiting {}s for item {} to cease existence".format(
                    timeout_secs, id
                )
            )

            def item_not_exists():
                try:
                    self.get(id)
                    return False
                except APIItemNotFoundException:
                    return True

            try:
                polling.poll(
                    lambda: item_not_exists(),
                    step=10,
                    poll_forever=False,
                    timeout=timeout_secs,
                )
                return True
            except polling.TimeoutException:
                return False

        # if state is not empty return success when resource current state is
        # in desired state
        else:
            waiting_for_status = [s.name for s in status]

            _log.debug(
                "waiting {}s for item {} to have status in {}".format(
                    timeout_secs, id, waiting_for_status
                )
            )
            try:

                def get_status():
                    status = getattr(self.get(id), self.status_fieldname)
                    return status

                polling.poll(
                    lambda: (get_status() in waiting_for_status),
                    step=10,
                    poll_forever=False,
                    timeout=timeout_secs,
                )
                _log.debug(
                    "Found item {} with status in {}".format(
                        id, waiting_for_status
                    )
                )
                return True
            except polling.TimeoutException:
                _log.debug(
                    "Timed out waiting for {} to have status in {}".format(
                        id, waiting_for_status
                    )
                )
                return False


@six.add_metaclass(abc.ABCMeta)
class AbstractResource:
    """Base class for Resource class repreenting an API resource.

    The resource class contains properties mapping to attributes in the
    response.

    The implementing class is declared in the ResourceController:

    Example
    -------
    class K8sClusterController(AbstractResourceController):
        ...
        resource_class = K8sCluster
    """

    def _get_all_fields(self):
        return self.all_fields

    def _set_all_fields(self, fields):
        self.all_fields = fields

    all_fields = abc.abstractproperty(_get_all_fields, _set_all_fields)

    def __init__(self, json):
        """Create a new Resource class.

        Parameters
        ----------
        json : obj
            JSON returned from the API for the Resource.
        """
        self.json = json

    def __repr__(self):
        """Return a represenation of Resource class."""
        return "<{} id:{}>".format(self.__class__.__name__, self.id)

    def __str__(self):
        """Return a str representation of a Resource class."""
        return "{}(id={})".format(self.__class__.__name__, self.id)

    @property
    def id(self):
        """@Field: from json['_links']['self']['href']."""
        return self.json["_links"]["self"]["href"]

    @property
    def _links(self):
        """@Field: from json['_links']."""
        return self.json["_links"]

    def __len__(self):
        """Return the number of resource fields in the Resource class."""
        return len(dir(self))


class ResourceList:
    """List of Resource objects."""

    def __init__(self, resource_class, json):
        """Create a list of resources using the resource_class.

        Parameters
        ----------
        resource_class : class
            Resource implementation class
        json : obj
            JSON return from the API
        """
        self.json = json
        self.resource_class = resource_class
        self.resources = [self.resource_class(j) for j in json]

    def __getitem__(self, item):
        """Retrieve a field value."""
        return self.resources[item]

    def tabulate(self, columns=[], style="pretty", display_headers=True):
        """Return a tabule output of the ResourceList.

        Parameters
        ----------
        columns : list, optional
            List of columns to output. The default value of an empty list
            will output all the available fields
        style : str, optional
            Table styles, by default "pretty"

            The available styles are:

            "plain"
            "simple"
            "github"
            "grid"
            "fancy_grid"
            "pipe"
            "orgtbl"
            "jira"
            "presto"
            "pretty"
            "psql"
            "rst"
            "mediawiki"
            "moinmoin"
            "youtrack"
            "html"
            "latex"
            "latex_raw"
            "latex_booktabs"
            "textile"

            See section 'Table Format' in https://pypi.org/project/tabulate/
            for more information

        Returns
        -------
        str
            table output of Resource

        Example
        -------
        Print the cluster list with all of the avaialble fields
        >>> print(hpeclient.cluster.list().tabulate())

        Print the cluster list with a subset of the fields
        >>> print(hpeclient.cluster.list().tabulate(
        ...     columns=['id', 'name','description']))
        """
        assert isinstance(columns, list), "'columns' parameter must be list"

        if len(columns) == 0:
            columns = self.resource_class.all_fields

        for field in columns:
            assert (
                field in self.resource_class.all_fields
            ), "item '{}' is not a field in {}.all_fields".format(
                field, self.__class__.__name__
            )

        self.display_fields = columns

        table = []
        for resource in self.resources:
            row = []
            for col in columns:
                if not hasattr(resource, col):
                    _log.warn(
                        "Field {} not found in {} - json {}".format(
                            col, resource, self.json
                        )
                    )
                row.append(getattr(resource, col, ""))
            table.append(row)

        if display_headers:
            output = tabulate(table, headers=columns, tablefmt=style)
        else:
            output = tabulate(table, tablefmt=style)

        if six.PY2:
            return output.encode(encoding="UTF-8", errors="strict")
        else:
            return output



# (C) Copyright [2020] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from __future__ import absolute_import

from .base_resource import AbstractWaitableResourceController, AbstractResource

from enum import Enum
from requests.structures import CaseInsensitiveDict

try:
    basestring
except NameError:
    basestring = str


class WorkerK8sStatus(Enum):
    bundle = 1
    installing = 2
    installed = 3
    ready = 4
    unlicensed = 5
    configuring = 6
    configured = 7
    error = 8
    sysinfo = 9
    unconfiguring = 10
    deleting = 11
    storage_pending = 12
    storage_configuring = 13
    storage_error = 14


class WorkerK8s(AbstractResource):

    all_fields = [
        "id",
        "status",
        "hostname",
        "ipaddr",
        "href",
        "_links",
    ]
    """All of the fields of a K8s Cluster objects that are returned by the HPE
    Container Platform API"""

    default_display_fields = [
        "id",
        "status",
        "hostname",
        "ipaddr",
    ]

    @property
    def worker_id(self):
        return int(self.json["_links"]["self"]["href"].split("/")[-1])

    @property
    def status(self):
        return self.json["status"]

    @property
    def hostname(self):
        return self.json["hostname"]

    @property
    def ipaddr(self):
        return self.json["ipaddr"]

    @property
    def href(self):
        return self.json["_links"]["self"]["href"]


class K8sWorkerController(AbstractWaitableResourceController):

    base_resource_path = "/api/v2/worker/k8shost"

    resource_list_path = "k8shosts"

    resource_class = WorkerK8s

    status_class = WorkerK8sStatus

    status_fieldname = "status"

    # TODO implement me!
    # def create_with_ssh_password(self, username, password):
    #     """Not Implemented yet"""
    #     raise NotImplementedError()

    def create_with_ssh_key(self, ip, ssh_key_data, tags=[]):
        """Create a gateway instance using SSH key credentials to access the host.

        Parameters
        ----------
        ip: str
            The IP address of the proxy host.  Used for internal
            communication.
        ssh_key_data: str
            The ssh key data as a string.
        tags: list
            Tags to use, e.g. "{ 'tag1': 'foo', 'tag2', 'bar' }".

        Returns
        -------
        string
            Worker ID
        """

        assert isinstance(
            ip, basestring
        ), "'ip' must be provided and must be a string"
        assert isinstance(
            ssh_key_data, basestring
        ), "'ssh_key_data' must be provided and must be a string"

        data = {
            "ipaddr": ip,
            "credentials": {
                "type": "ssh_key_access",
                "ssh_key_data": ssh_key_data,
            },
            "tags": tags,
        }

        response = self.client._request(
            url="/api/v2/worker/k8shost/",
            http_method="post",
            data=data,
            description="worker/create_with_ssh_key",
        )
        return CaseInsensitiveDict(response.headers)["location"]

    def get(self, id, setup_log=False):

        if setup_log is True:
            params = {"setup_log": "true"}
        else:
            params = {}

        return super(K8sWorkerController, self).get(id, params)

    def set_storage(self, worker_id, ephemeral_disks=[], persistent_disks=[]):
        """Set storage for a k8s worker.

        Parameters
        ----------
        worker_id : str
            The k8s worker ID, format - '/api/v2/worker/k8shost/[0-9]+'
        ephemeral_disks : list
            List of ephemeral disks. Mandatory parameter, by default []
        persistent_disks : list, optional
            List of persistent disks, by default []

        Raises
        ------
        APIItemNotFoundException
        APIItemConflictException
        APIException
        AssertionError
        """
        # Make sure that the worker exists
        self.get(worker_id)

        assert isinstance(
            ephemeral_disks, list
        ), "'ephemeral_disks' must be provided and and must be a list"

        assert (
            len(ephemeral_disks) > 0
        ), "'ephemeral_disks' must contain at least one disk"

        assert isinstance(
            persistent_disks, list
        ), "'persistent_disks' must be a list"

        # Prepare the payload
        data = {
            "op_spec": {
                "ephemeral_disks": ephemeral_disks,
                "persistent_disks": persistent_disks,
            },
            "op": "storage",
        }

        # Make the request
        self.client._request(
            url=worker_id,
            http_method="post",
            data=data,
            description="worker/set_storage",
        )



# (C) Copyright [2020] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from __future__ import absolute_import

from .base_resource import AbstractWaitableResourceController, AbstractResource

from enum import Enum
import re
from requests.structures import CaseInsensitiveDict

try:
    basestring
except NameError:
    basestring = str


class K8sClusterStatus(Enum):
    """Bases: enum.Enum

    The statuses for a K8S Cluster

    **Note:**

    The integer values do not have a meaning outside of this library.
    The API uses a string identifier with the status name rather than an
    integer value.
    """

    ready = 1
    creating = 2
    updating = 3
    upgrading = 4
    deleting = 5
    error = 6
    warning = 7


class K8sCluster(AbstractResource):
    """Create an instance of K8sCluster from json data returned from the HPE
    Container Platform API.

    Users of this library are not expected to create an instance of this
    class.

    Parameters
    ----------
    json : str
        The json returned by the API representing a K8sCluster.

    Returns
    -------
    K8sCluster:
        An instance of K8sCluster
    """

    all_fields = [
        "id",
        "name",
        "description",
        "k8s_version",
        "addons",
        "created_by_user_id",
        "created_by_user_name",
        "created_time",
        "k8shosts_config",
        "admin_kube_config",
        "dashboard_token",
        "api_endpoint_access",
        "dashboard_endpoint_access",
        "cert_data",
        "status",
        "status_message",
        "_links",
    ]
    """All of the fields of a K8s Cluster objects that are returned by the HPE
    Container Platform API"""

    default_display_fields = [
        "id",
        "name",
        "description",
        "k8s_version",
        "status",
    ]

    @property
    def name(self):
        """@Field: from json['label']['name']"""
        return self.json["label"]["name"]

    @property
    def description(self):
        """@Field: from json['label']['description']"""
        return self.json["label"]["description"]

    @property
    def k8s_version(self):
        """@Field: from json['k8s_version']"""
        return self.json["k8s_version"]

    @property
    def addons(self):
        """@Field: from json['addons']"""
        if "addons" in self.json:
            return self.json["addons"]
        else:
            return ""

    @property
    def created_by_user_id(self):
        """@Field: from json['created_by_user_id']"""
        return self.json["created_by_user_id"]

    @property
    def created_by_user_name(self):
        """@Field: from json['created_by_user_name']"""
        return self.json["created_by_user_name"]

    @property
    def created_time(self):
        """@Field: from json['created_time']"""
        return self.json["created_time"]

    @property
    def k8shosts_config(self):
        """@Field: from json['k8shosts_config']"""
        return self.json["k8shosts_config"]

    @property
    def admin_kube_config(self):
        """@Field: from json['admin_kube_config']"""
        if "admin_kube_config" in self.json:
            return self.json["admin_kube_config"]
        else:
            return ""

    @property
    def dashboard_token(self):
        """@Field: from json['dashboard_token']"""
        if "dashboard_token" in self.json:
            return self.json["dashboard_token"]
        else:
            return ""

    @property
    def api_endpoint_access(self):
        """@Field: from json['api_endpoint_access']"""
        if "api_endpoint_access" in self.json:
            return self.json["api_endpoint_access"]
        else:
            return ""

    @property
    def dashboard_endpoint_access(self):
        """@Field: from json['dashboard_endpoint_access']"""
        if "dashboard_endpoint_access" in self.json:
            return self.json["dashboard_endpoint_access"]
        else:
            return ""

    @property
    def cert_data(self):
        """@Field: from json['cert_data'] or None if cert_data not available"""
        try:
            return self.json["cert_data"]
        except KeyError:
            return None

    @property
    def status(self):
        """@Field: from json['status']"""
        return self.json["status"]

    @property
    def status_message(self):
        """@Field: from json['status_message']"""
        if "status_message" in self.json:
            return self.json["status_message"]
        else:
            return ""


class K8sClusterHostConfig:
    """Object to represent a pair of `host node` and the `role` of the host
    - `master` or `worker`.
    """

    @classmethod
    def create_from_list(cls, noderole):
        """Factory method to create K8sClusterHostConfig from a list.

        Parameters
        ----------
        noderole: list
            the noderole must only have two values: [ node, role ]

        See Also
        --------
        See :py:meth:`K8sClusterHostConfig` for the allowed node and role
        values.
        """

        assert (
            len(noderole) == 2
        ), "'noderole' list must have two values [ node, role ]"
        return K8sClusterHostConfig(node=noderole[0], role=noderole[1])

    def __init__(self, node, role):
        assert isinstance(node, basestring), "'node' must be an string"
        assert re.match(
            r"\/api\/v2\/worker\/k8shost\/[0-9]+", node
        ), "'node' must have format '/api/v2/worker/k8shost/[0-9]+'"
        assert role in [
            "master",
            "worker",
        ], "'role' must one of ['master, worker']"

        self.node = node
        self.role = role

    def to_dict(self):
        """Returns a dict representation of the object.

        Returns
        -------
        dict

        Example
        -------
        >>> .to_dict()
        {
            'node': '/api/v2/worker/k8shost/12',
            'role': 'master'
        }
        """
        return {"node": self.node, "role": self.role}


class K8sClusterController(AbstractWaitableResourceController):
    """Class for interacting with K8S Clusters.

    An instance of this class is available in the
    client.ContainerPlatformClient with the attribute name
    :py:attr:`k8s_cluster <.client.ContainerPlatformClient.k8s_cluster>`.  The
    methods of this class can be invoked using `client.k8s_cluster.method()`.
    See the example below:

    Example
    -------
    >>> client = ContainerPlatformClient(...).create_session()
    >>> client.k8s_cluster.list()
    """

    base_resource_path = "/api/v2/k8scluster"

    resource_list_path = "k8sclusters"

    resource_class = K8sCluster

    status_class = K8sClusterStatus

    status_fieldname = "status"

    def create(
        self,
        name=None,
        description=None,
        k8s_version=None,
        pod_network_range="10.192.0.0/12",
        service_network_range="10.96.0.0/12",
        pod_dns_domain="cluster.local",
        persistent_storage_local=False,
        persistent_storage_nimble_csi=False,
        k8shosts_config=[],
        addons=[],
    ):
        """Send an API request to create a K8s Cluster.  The cluster creation
        will be asynchronous - use the :py:meth:`wait_for_status` method to
        wait for the cluster to be created.

        For the list of possible statuses see :py:class:`K8sClusterStatus`.

        Parameters
        ----------
        name: str
            Cluster name - required parameter.  Name must be at least 1
            character
        description: str
            Cluster description - defaults to empty string if not provided
        k8s_version: str
            Kubernetes version to configure. If not specified defaults to
            the latest version as supported by the rpms.
        pod_network_range: str
            Network range to be used for kubernetes pods. Defaults to
            `10.192.0.0/12`
        addons: list
            Addons - See :py:method:`get_available_addons`.
        service_network_range: str
            Network range to be used for kubernetes services that are
            exposed with Cluster IP. Defaults to `10.96.0.0/12`
        pod_dns_domain: str
            DNS Domain to be used for kubernetes pods. Defaults to
            `cluster.local`
        persistent_storage_local: str
            Enables local host storage to be available in the kubernetes
            cluster
        persistent_storage_nimble_csi: bool
            Set to True to installs the Nimble CSI plugin for Nimble
            storage to be available in the kubernetes cluster
        k8shosts_config: list[K8sClusterHostConfig]
            list of :py:class:`K8sClusterHostConfig` objects determining
            which hosts to add and their role (master or worker)

        Returns
        -------
        str
            K8s Cluster ID with the format: '/api/v2/k8scluster/[0-9]+'

        Raises
        ------
        APIException
        """

        assert (
            isinstance(name, basestring) and len(name) > 0
        ), "'name' must be provided and must be a string"
        assert description is None or isinstance(
            description, basestring
        ), "'description' if provided, must be a string"
        assert k8s_version is None or isinstance(
            k8s_version, basestring
        ), "'k8s_version' if provided, must be a string"
        assert isinstance(
            pod_network_range, basestring
        ), "'pod_network_range' must be a string"
        assert isinstance(
            service_network_range, basestring
        ), "'service_network_range' must be a string"
        assert isinstance(
            pod_dns_domain, basestring
        ), "'pod_dns_domain' must be a string"
        assert isinstance(
            persistent_storage_local, bool
        ), "'persistent_storage_local' must be True or False"
        assert isinstance(
            persistent_storage_nimble_csi, bool
        ), "'persistent_storage_nimble_csi' must be True or False"
        assert isinstance(
            k8shosts_config, list
        ), "'k8shosts_config' must be a list"
        assert (
            len(k8shosts_config) > 0
        ), "'k8shosts_config' must have at least one item"
        for i, conf in enumerate(k8shosts_config):
            assert isinstance(conf, K8sClusterHostConfig), (
                "'k8shosts_config' item '{}' is not of"
                " type K8sClusterHostConfig"
            ).format(i)
        assert isinstance(addons, list), "'addons' must be a list"

        data = {
            "label": {"name": name},
            "pod_network_range": pod_network_range,
            "service_network_range": service_network_range,
            "pod_dns_domain": pod_dns_domain,
            "addons": addons,
            "persistent_storage": {
                "local": persistent_storage_local,
                "nimble_csi": persistent_storage_nimble_csi,
            },
            "k8shosts_config": [c.to_dict() for c in k8shosts_config],
        }
        if description is not None:
            data["label"]["description"] = description
        if k8s_version is not None:
            data["k8s_version"] = k8s_version

        response = self.client._request(
            url="/api/v2/k8scluster",
            http_method="post",
            data=data,
            description="k8s_cluster/create",
        )
        return CaseInsensitiveDict(response.headers)["Location"]

    def get(self, id, params={}, setup_log=False):
        """Retrieve a K8s Cluster.

        Parameters
        ----------
        id: str
            The k8s cluster ID
        """
        if setup_log is True:
            params["setup_log"] = "true"

        return super(K8sClusterController, self).get(id=id, params=params)

    def k8smanifest(self):
        """Retrieve the k8smanifest.

        Returns
        -------
        json
            K8sManifest

        Raises
        ------
        APIException
        """
        response = self.client._request(
            url="/api/v2/k8smanifest",
            http_method="get",
            description="k8s_cluster/k8smanifest",
        )
        return response.json()

    def k8s_supported_versions(self):
        """Retrieve list of K8S Supported Versions.

        Returns
        -------
        list[str]
            List of K8s Supported Versions

        Raises
        ------
        APIException
        """
        return self.k8smanifest()["supported_versions"]

    def get_available_addons(self, id=None, k8s_version=None):
        """Retrieve list of K8S Supported Versions.

        Parameters
        ----------
        id: str
            The k8s cluster ID

        Returns
        -------
        list[str]
            List of available addons

        Raises
        ------
        APIException
        """
        assert (
            id is not None or k8s_version is not None
        ), "Either 'id' or 'k8s_version' parameter must be provided"
        assert (
            id is None or k8s_version is None
        ), "Either 'id' or 'k8s_version' parameter must be provided"

        if id:
            k8s_version = self.get(id).k8s_version

        return self.k8smanifest()["version_info"][k8s_version]["addons"]

    def add_addons(self, id, addons=[]):
        """Retrieve list of K8S Supported Versions.

        Parameters
        ----------
        id: str
            The k8s cluster ID
        addons: list
            The list of addons to add.

        Raises
        ------
        APIException
        """
        # TODO assert ID is provided and valid
        assert (
            isinstance(addons, list) and len(addons) > 0
        ), "'Addons' parameter must be a list and have at least one entry."

        current_addons = self.get(id).addons
        required_addons = current_addons + addons

        # de-duplicate
        required_addons = list(dict.fromkeys(required_addons))

        data = {
            "change_spec": {"addons": required_addons},
            "operation": "reconfigure",
            "reason": "",
        }

        self.client._request(
            url="{}/change_task".format(id),
            http_method="post",
            description="k8s_cluster/add_addons",
            data=data,
        )
